/*----------------------------------------------------------------------
 *  TemplateNode
 *----------------------------------------------------------------------
 */
 
function TemplateNode(sampleADN, children) {
    this.sampleADN = sampleADN;
    this.children = children;
    this.optional = false;
    this.repeating = false;
}

TemplateNode.createTemplateNode = function(adn) {
    var children = map(
        function(adnNode) {
            return new TemplateNode(adnNode, null);
        },
        adn.children
    );
    return new TemplateNode(adn, children);
}

TemplateNode.prototype.fillChildren = function(adn) {
    this.children = map(
        function(adnNode) {
            return new TemplateNode(adnNode, null);
        },
        adn.children
    );
}

TemplateNode.prototype.learnChildren = function(adn) {
    if (this.children == null) {
        throw new Error("Internal error");
    }
    
    var costFunction = function(templateNode, adn) {
        return templateNode.sampleADN.computeEditCostIncludeChildren(adn);
    };
    var compareFunction = function(templateNode, adn) {
        return (templateNode.sampleADN.nodeName != adn.nodeName) ? 1 : 0;
    };
    var markOptionalFunction = function(templateNode) {
        templateNode.optional = true;
    };
    var constructorFunction = function(adn) {
        return new TemplateNode(adn, null);
    };
    
    Template._learn(
        this.children, 
        adn.children, 
        costFunction, 
        compareFunction, 
        markOptionalFunction, 
        constructorFunction
    );
}

TemplateNode.prototype.bindChildren = function(adn) {
    /*
     *  The trick here is that the template should always
     *  has "more" children than the adn (considering
     *  wildcards).
     */
     
    var us = 0;
    
    for (var them = 0; them < adn.children.length; them++) {
        var theirADN = adn.children[them];
        var match = false;
        
        var bestOffset = 0;
        var minCost = 1000;
        
        for (var i = us; i < this.children.length; i++) {
            var ourTemplateNode = this.children[i];
            var ourADN = ourTemplateNode.sampleADN;
            
            var c = theirADN.computeEditCostIncludeChildren(ourADN);
            if (ourTemplateNode.optional) {
                c = c / (1 + i - us);
            }
            
            if (c < minCost) {
                bestOffset = i - us;
                minCost = c;
            }
            
            if (!ourTemplateNode.optional) {
                break;
            }
        }
        
        us += bestOffset;
        
        var binding = {
            templateNode : this.children[us],
            templateNodeIndex : us
        };
        
        theirADN.binding = binding;
        us++;
    }
}

TemplateNode.prototype.learnText = function(adns) {
    this.children = [];
    
    var i = 0;
    while (i < adns.length) {
        if (adns[i]) break;
        i++;
    }
    if (i >= adns.length) {
        return;
    }
    
    var textTemplateNodes = map(
        function(token) { 
            return new TextTemplateNode([ token ]); 
        }, 
        tokenize(adns[i].text)
    );
    
    var learn = function(templates, samples, recurse) {
        if (templates.length == 0) {
            return map(
                function(token) { 
                    var node = new TextTemplateNode([ token ]);
                    node.variable = true;
                    return node;
                }, samples
            );
        } else if (samples.length == 0) {
            map(
                function(templateNode) { 
                    templateNode.variable = true;
                    return templateNode;
                }, templates
            );
            return templates;
        }
        
        var bestSampleIndex = -1;
        var bestTemplateIndex = -1;
        var bestTokenCount = -1;
        var bestRunningLength = -1;
        
        for (var i = 0; i < samples.length; i++) {
            var sample = samples[i];
            
            for (var j = 0; j < templates.length; j++) {
                var template = templates[j];
                if (sample.text == template.tokens[0].text) {
                    var runningLength = 0;
                    var tokenCount = 0;
                    
                    for (var k = j; k < templates.length && (i + (k - j)) < samples.length; k++) {
                        var template2 = templates[k];
                        var sample2 = samples[i + (k - j)];
                        //if (sample2.text == template2.tokens[0].text) {
                        if (sample2.type == template2.tokens[0].type &&
                            (sample2.type != 0 || sample2.text == template2.tokens[0].text)) {
                            runningLength += sample2.text.length;
                            tokenCount++;
                        } else {
                            break;
                        }
                    }
                    
                    if (runningLength > bestRunningLength) {
                        bestSampleIndex = i;
                        bestTemplateIndex = j;
                        bestTokenCount = tokenCount;
                        bestRunningLength = runningLength;
                    }
                }
            }
        }
        
        if (bestSampleIndex >= 0) {
            var leftTemplates = templates.slice(0, bestTemplateIndex);
            var midTemplates = templates.slice(bestTemplateIndex, bestTemplateIndex + bestTokenCount);
            var rightTemplates = templates.slice(bestTemplateIndex + bestTokenCount);
            
            var leftSamples = samples.slice(0, bestSampleIndex);
            var midSamples = samples.slice(bestSampleIndex, bestSampleIndex + bestTokenCount);
            var rightSamples = samples.slice(bestSampleIndex + bestTokenCount);
            
            for (var i = 0; i < midTemplates.length; i++) {
                if (midTemplates[i].tokens[0].text != midSamples[i].text) {
                    midTemplates[i].variable = true;
                }
            }
            
            return recurse(leftTemplates, leftSamples, recurse).
                concat(midTemplates).
                concat(recurse(rightTemplates, rightSamples, recurse));
        } else {
            map(
                function(templateNode) { 
                    templateNode.variable = true;
                    return templateNode;
                }, templates
            );
            return templates;
        }
    };
    
    for (var j = i + 1; j < adns.length; j++) {
        if (adns[j]) {
            textTemplateNodes = learn(textTemplateNodes, tokenize(trim(adns[j].text)), learn);
        }
    }
    
    /*
     *  Consolidate consecutive tokens
     */
    var i = 0;
    while (i < textTemplateNodes.length) {
        var type = textTemplateNodes[i].tokens[0].type;
        var variable = textTemplateNodes[i].variable;
        
        var tokens = [];
        while (i < textTemplateNodes.length && 
                textTemplateNodes[i].variable == variable && 
                textTemplateNodes[i].tokens[0].type == type) {
                
            tokens.push(textTemplateNodes[i].tokens[0]);
            i++;
            
            if (type != 0) {
                break;
            }
        }
        var newTextTemplateNode = new TextTemplateNode(tokens);
        newTextTemplateNode.variable = variable;
        
        this.children.push(newTextTemplateNode);
    }
}

TemplateNode.prototype.bindText = function(adn) {
    var bindings = [];
    var newTextTemplateNodes = this.children;
    
    var tokens = tokenize(adn.text);
    for (var i = 0; i < newTextTemplateNodes.length; i++) {
        var textTemplateNode = newTextTemplateNodes[i];
        var binding = null;
        
        if (tokens.length > 0) {
            if (textTemplateNode.variable) {
                if (textTemplateNode.tokens[0].type == 0) { // string
                    /*
                     *  A variable template node of strings don't have any 
                     *  fixed count of tokens. We need to find the next
                     *  template token, if any.
                     */
                    if (i + 1 < newTextTemplateNodes.length) {
                        var nextTextTemplateNodes = newTextTemplateNodes[i + 1];
                        var k = 1;
                        if (nextTextTemplateNodes.variable) {
                            while (k < tokens.length) {
                                if (tokens[k].type != 0) {
                                    break;
                                }
                                k++;
                            }
                        } else {
                            while (k < tokens.length) {
                                if (tokens[k].type != 0 || 
                                    tokens[k].text == nextTextTemplateNodes.tokens[0].text) {
                                    break;
                                }
                                k++;
                            }
                        }
                        binding = {
                            text: map(function(t) { return t.text; }, tokens.slice(0, k)).join(" ")
                        };
                        tokens = tokens.slice(k);
                    } else {
                        /*
                         *  All subsequent tokens are variable.
                         */
                        binding = {
                            text: map(function(t) { return t.text; }, tokens).join(" ")
                        };
                        tokens = null;
                    }
                } else {
                    if (tokens[0].type == textTemplateNode.tokens[0].type) {
                        binding = {
                            text: tokens[0].text
                        };
                        tokens = tokens.slice(1);
                    }
                }
            } else {
                var k = 0;
                while (k < tokens.length && k < textTemplateNode.tokens.length) {
                    if (tokens[k].type != textTemplateNode.tokens[k].type) {
                        break;
                    }
                    k++;
                }
                
                binding = {
                    text: map(function(t) { return t.text; }, tokens.slice(0, k)).join(" ")
                }
                tokens = tokens.slice(k);
            }
        }
        bindings.push(binding);
    }
    
    return bindings;
}

/*----------------------------------------------------------------------
 *  TextTemplateNode
 *----------------------------------------------------------------------
 */
 
function TextTemplateNode(tokens) {
    this.tokens = tokens;
    this.variable = false;
}
 
/*----------------------------------------------------------------------
 *  Template
 *----------------------------------------------------------------------
 */
 
function Template(templateNodes) {
    this.children = templateNodes;
}

Template.prototype.learn = function(item) {
    for (var i = 0; i < item.children.length; i++) {
        this.children[i].learnChildren(item.children[i]);
    }
}

Template.prototype.bind = function(item) {
    for (var i = 0; i < item.children.length; i++) {
        this.children[i].bindChildren(item.children[i]);
    }
}


Template._learn = function(templateItems, sampleItems, costFunction, compareFunction, markOptionalFunction, constructorFunction) {
    var us = 0;
    var them = 0;
    
    while (us < templateItems.length && them < sampleItems.length) {
        var templateItem = templateItems[us];
        var sampleItem = sampleItems[them];
        
        /*
         *  Find their best that matches ours
         */
        var theirBestOffset = 0;
        var theirMinCost = 1000;
        for (var i = them; i < sampleItems.length; i++) {
            var c = costFunction(templateItem, sampleItems[i]) * (theirBestOffset + 1);
            if (c < theirMinCost) {
                theirBestOffset = i - them;
                theirMinCost = c;
            }
        }
        
        /*
         *  Find our best that matches theirs
         */
        var ourBestOffset = 0;
        var ourMinCost = 1000;
        for (var i = us; i < templateItems.length; i++) {
            var c = costFunction(templateItems[i], sampleItem) * (ourBestOffset + 1);
            if (c < ourMinCost) {
                ourBestOffset = i - us;
                ourMinCost = c;
            }
        }
        
        if (theirMinCost < ourMinCost) {
            if (compareFunction(templateItem, sampleItems[them + theirBestOffset]) != 0) {
                markOptionalFunction(templateItem);
            }
            
            /*
             *  Insert new template nodes from them and mark optional
             */
            var args = [ us, 0 ];
            for (var offset = 0; offset < theirBestOffset; offset++) {
                var templateItem = constructorFunction(sampleItems[them + offset]);
                markOptionalFunction(templateItem);
                
                args.push(templateItem);
            }
            templateItems.splice.apply(templateItems, args);
            
            us += theirBestOffset + 1; // the insertion affects our offset, too
            them += theirBestOffset + 1;
        } else {
            if (compareFunction(templateItems[us + ourBestOffset], sampleItem) != 0) {
                markOptionalFunction(templateItems[us + ourBestOffset]);
            }
            
            /*
             *  Mark some of our nodes as optional
             */
            for (var offset = 0; offset < ourBestOffset; offset++) {
                markOptionalFunction(templateItems[us + offset]);
            }
            
            us += ourBestOffset + 1;
            them += 1;
        }
    }
    
    if (us < templateItems.length) {
        /*
         *  The rest of our nodes are optional
         */
        while (us < templateItems.length) {
            markOptionalFunction(templateItems[us]);
            us++;
        }
    } else if (them < sampleItems.length) {
        /*
         *  Insert new template nodes from them and mark optional
         */
        while (them < sampleItems.length) {
            var templateItem = constructorFunction(sampleItems[them]);
            markOptionalFunction(templateItem);
            
            templateItems.push(templateItem);
            
            them++;
        }
    }
}
